/* LDDL - Little Data Description Language
 *
 *	Copyright (c) 2000,  Marty Olevitch and
 *		Cosmic Ray Group, Washington University in St. Louis
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the "Artistic License", the terms of
 *	which are found in the file "Artistic".
 *
 *	Marty Olevitch, Winter 1999-2000
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>

#include "list.h"
#include "misc.h"
#include "args.h"
#include "var.h"

#include "ds.h"
#include "lddl.h"
#include "y.tab.h"

static char *cvsid = "$Id: lddl.c,v 1.1.1.1 2000/03/24 20:10:24 marty Exp $";

extern FILE *yyin;

#ifdef USE_INT_FOR_LONG
/* dec alpha long is 8 bytes, but int is 4 */
#define long int
#define u_long int
#endif

static int Debug = 0;
static char *Progname;

static char **Argvp;			/* scan pointer for nextfile() */
static char *Nullargv[] = { "-", NULL };/* dummy argv for case of no args */
static char *Inname;			/* filename for messages etc. */
static u_long Bytecnt;			/* byte count for messages etc. */
static FILE *In = NULL;			/* current input file */

static FILE *Sfp;	/* file pointer for format file */

/* pointers to application callback functions */
static evt_cb _pre_evtf = NULL;
static evt_cb _evtf = NULL;
static evt_cb _post_evtf = NULL;
static group_cb _groupf = NULL;
static group_cb _post_groupf = NULL;
static node_cb _nodef = NULL;
static void (*_app_errf)(char *str) = NULL;

/* evt struct has more details about each particular kind of event. These
 * are the elements of the list identified by Evt_kind_id. */
struct evt {
    int cond_id;	/* list of cond structs - identifies evt kind */
    int node_id;	/* list of node structs - elements of evt */
    char *name;		/* name of this event kind */
    int argc;		/* arg count for applications */
    char **argv;	/* arg vector for applications */
};
typedef struct evt evt;

/* evt_kind_cond struct has information on each condition that is used to
 * identify an event kind in the data stream. */
struct evt_kind_cond {
    int offset;		/* byte offset from evt start to evt kind element */
    int type;		/* type of evt kind element (LDDL_INT8,INT16,or INT32) */
    valstruct mask;	/* bitmask, if any, for condition value */
    valstruct val;	/* value for the condition to be true */
};
typedef struct evt_kind_cond evt_kind_cond;

/* The "node" describes an element in a the data. It corresponds to the
 * "typespec" token in grammar.y. This structure is the element contained
 * on the various linked lists of nodes. */
struct node {
    int type;		/* type of data item (LDDL_INT8, etc from grammar.y) */
    int cnt;		/* number of data items */
    int nbytes;		/* no. of bytes in cnt data items */
    char *nbytes_var;	/* var. name from which to get nbytes */
    char *value_to_var;	/* var. name to which to set value of this node */
    int optional;	/* not 0 if node with this value need not be present */
    int group_id;	/* used for "groups" if type == LDDL_GROUP */
    char *group_name;	/* name of the group */
    valstruct value;	/* values that node must match; NULL if none. */
    valstruct mask;	/* bitmask to "and" with value; NULL if none. */
    int argc;		/* arg count for applications */
    char **argv;	/* arg vector for applications */
};
typedef struct node node;

/* Charcnt - no. bytes processed so far */
static unsigned long Charcnt = 0L;

/* groupnode - information on each "group" */
struct groupnode {
    char *name;		/* group name */
    int id;		/* id of list of nodes comprising this "group" */
};
typedef struct groupnode groupnode;

static int Group_list_id;	/* groupnode list id */
static int Cur_group_id = -1;	/* currently accumulating group node list */
static char *Cur_group_name;	/* name of currently accumulating group */

/* Evt_kind_rdamt - no. of bytes to look at to determine event kind.
 * Calculated by the program based on each cond's offset and type. */
static int Evt_kind_rdamt = 0;

/* Evt_kind_id - linked list id for chain of evt structs */
static int Evt_kind_id;

#define DEF_EVT_NAME	"$default"	/* default evt name */
static char *Cur_evt_name;		/* name of current evt kind */
static int Check_evt_kind;		/* nonzero to check for evt kind */

static int Cur_cond_id;	/* id of cond list currently in use */
static int Cur_node_id;	/* id of node list currently in use */

/* Endian stuff - this is for determining if the data is in our system's
 * native endianness from the data itself. */
int Check_endianness = 0;	/* overridden by Native_set */
int Endian_offset = 0;		/* byte offset to endian marker */
int Endian_rdamt;		/* no. bytes to read to get to endian marker */
int Endian_type;		/* type of endian marker (INT16 or INT32) */
valstruct Endian_val;		/* value to check for */
valstruct Endian_mask;		/* ...possibly anded by this value */
static int endian_determine(void);

#define DEF_BUF_SIZE	32768		/* default data buffer size */

static unsigned char *Inp;
static unsigned char *Swapp;

static int check_enough_data(int atleast);
static int do_node_list(int id, char *name);
static int do_normal_node(node *np);
static node *fill_node(int type, int cnt, char *nbytes_var,
    char *value_to_var, int optional, valstruct *v, int mask,
    int group_id, char *group_name);
static size_t getbuf(char *ptr, size_t size, size_t rdamt);
static int match_evt_kind(evt_kind_cond *cp, u_char *p);
static char *nextfile(void);
static void print_node(node *np);
static void print_node_list(int id);
static void process(void);
static void revbytes(u_char *srcp, u_char *destp, int width, int cnt);

/* Force_arg[cv] - if a group is encountered with argc set, then we set
 * argc and argv for every node in the group and unset after the group has
 * been processed. Currently, Force_arg[cv] does not override an argv that
 * was already set for a node nor a forced argv that was set in a group
 * that includes the current group (I think). */
static int Force = 0;
static int Force_argc;
static char **Force_argv;

/* Native set by lddl_set_native() - overrides Check_endianness */
static int Native_set = 0;	
static int Native = 1;		/* if !0, don't use swapped values in data */

/* lddl_start - parse format description file, read data */
int
lddl_start(char *fmtfile, int ninfiles, char **infilev, char *progname)
{
    evt *ep;
    int ret;

    Progname = progname;

    if (fmtfile == NULL) {
	fatal("%s: no data format spec file!", Progname);
    }

    if (Debug) {
	fprintf(stderr, "fmtfile: %s\n", fmtfile);
    }

    if (var_init()) {
	fatal("%s: bad var_init", Progname);
    }

    Evt_kind_id = list_new();
    if (Evt_kind_id < 0) {
	fatal("%s: bad list_new on main evt list", Progname);
    }

    /* Set up the default node list. */
    Cur_evt_name = str_alloc(NULL, DEF_EVT_NAME);
    strcpy(Cur_evt_name, DEF_EVT_NAME);

    evt_kind_init(Cur_evt_name);
    if (list_first(Evt_kind_id)) {
	fatal("%s: bad list_first default node list", Progname);
    }
    if (list_val(Evt_kind_id, (void **)&ep) <= 0) {
	fatal("%s: bad list_val default node list", Progname);
    }
    Cur_node_id = ep->node_id;
    Check_evt_kind = 0;

    Sfp = fopen(fmtfile, "r");
    if (Sfp == NULL) {
	fatal("%s: can't open format description file '%s' (%s)",
	    Progname, fmtfile, strerror(errno));
    }
    yyin = Sfp;

    if (yyparse()) {
	exit(1);
    }
    if (mitten_drin()) {
	/* end of input in the middle of a grammar construct */
	exit(1);
    }

    fclose(Sfp);

    if (ninfiles) {
	Argvp = infilev;
    } else {
	Argvp = Nullargv;
    }

    /*****
    if (optind >= argc) {
	Argvp = Nullargv;
    } else {
	Argvp = &argv[optind];
    }
    *****/

    Inname = nextfile();
    if (Inname != NULL) {
	process();
    }
}

/* getbuf - get next buffer of data*/
static size_t
getbuf(ptr, size, rdamt)
    char *ptr;
    size_t size;
    size_t rdamt;
{
    char *namep;
    int nrd;
    size_t totrd = (size_t)0;

    while ((nrd = fread(ptr, size, rdamt, In)) != (size*rdamt)) {
	if (ferror(In)) {
	    warning("%s: Error reading file '%s' at bytecnt=%ld (%s)", 
	    	Progname, Inname, Bytecnt, strerror(errno));
	}
	if (Debug && feof(In)) {
	    warning("%s: EOF on file '%s'", Progname, Inname);
	}
	namep = nextfile();
	if (namep == NULL) {
	    break;
	}
	Inname = namep;		/* only after we know it's good */
	rdamt -= nrd;
	totrd += nrd;
	ptr += (nrd*size);
    }
    totrd += nrd;
    Bytecnt += (totrd*size);
    return(totrd);
}

/* nextfile - switch files */
static char *	/* filename */
nextfile(void)
{
    char *namep;
    struct stat statbuf;

    if (In != NULL) {
	(void) fclose(In);
    }

    namep = *Argvp;
    if (namep == NULL) {	/* no more files */
	return(NULL);
    }
    Argvp++;

    if (STREQ(namep, "-")) {
	In = stdin;
	namep = "stdin";
    } else {
	In = fopen(namep, "r");
	if (In == NULL) {
	    fatal("%s: can't open input file '%s' (%s)", Progname,
	    	namep, strerror(errno));
	}
	if (fstat(fileno(In), &statbuf) < 0) {
	    fatal("%s: can't fstat `%s' (%s)", Progname, 
	    	namep, strerror(errno));
	}
	if ((statbuf.st_mode & S_IFMT) == S_IFDIR) {
	    fatal("%s: '%s' is  a directory!", Progname, namep);
	}
    }

    Bytecnt = 0;
    return(namep);
}

/* process - process input data */
static void
process(void)
{
    int cond;
    evt_kind_cond *cp;
    evt *ep;
    int match;

    if (Check_endianness) {
	/* figure out from the data stream if the data is in our system's
	 * native endianness. */
	if (endian_determine()) {
	    /* EOF */
	    return;
	}
    }

    while (1) {
	/* Set ep to point to the default event. This will be overridden if
	 * we are processing different kinds of events. */
	if (list_first(Evt_kind_id)) {
	    fatal("process: bad Evt_kind list_first");
	}
	(void)list_val(Evt_kind_id, (void **)&ep);

	if (_pre_evtf != NULL) {
	    /* before event function */
	    (void)_pre_evtf(NULL, 0, NULL);
	}

	/* If we are processing different kinds of events, then the
	 * following 'if' is entered. On successful match of evt kind, the
	 * Cur_node_id and Cur_evt_name will be set and passed to
	 * do_node_list() where the real work is handled. If we are just
	 * using the default event nodes ($default), then Cur_node_id and
	 * Cur_evt_name is already set, and passed to do_node_list(). */
	if (Check_evt_kind) {
	    /* attempt to match the evt kind */
	    if (check_enough_data(Evt_kind_rdamt)) {
		/* EOF */
		return;
	    }

	    while (list_val(Evt_kind_id, (void **)&ep) > 0) {
		match = 1;
		cond = ep->cond_id;
		if (list_first(cond)) {
		    fatal("process: bad cond list_first");
		}

		/* Check the conditions for this evt. All of the conditions
		 * for the evt must be true. */
		while (list_val(cond, (void **)&cp) > 0) {
		    if (!match_evt_kind(cp, Inp + cp->offset)) {
			match = 0;
			break;
		    }
		}

		if (match) {
		    break;
		}
	    }

	    if (!match) {
		fatal("process: can't match evt kind");
		exit(1);
	    }

	    Cur_node_id = ep->node_id;

	    Cur_evt_name = str_alloc(Cur_evt_name, ep->name);
	    strcpy(Cur_evt_name, ep->name);
	}

	if (_evtf != NULL) {
	    /* identified event function */
	    (void)_evtf(Cur_evt_name, ep->argc, ep->argv);
	}

	if (ep->argc && !Force) {
	    Force++;
	    Force_argc = ep->argc;
	    Force_argv = ep->argv;
	}

	if (do_node_list(Cur_node_id, Cur_evt_name)) {
	    /* EOF */
	    return;
	}

	if (ep->argc && !Force) {
	    Force--;
	}

	if (_post_evtf != NULL) {
	    /* post event function */
	    (void)_post_evtf(Cur_evt_name, ep->argc, ep->argv);
	}
    }
}

/* do_node_list - process the nodes on the list. May be called recursively to
 * deal with "groups". Returns 1 on EOF, else 0. */
static int
do_node_list(int id, char *name)
{
    int count;
    int i;
    node *np;
    int ret;

    if (list_first(id)) {
	fatal("process: bad list_first");
    }

#ifdef NOTDEF
    if (name != NULL) {
	fprintf(stderr, "+++++++++ %s (%0x)\n", name, Inp);
    } else {
	fprintf(stderr, "+++++++++ ???? (%0x)\n", Inp);
    }
#endif /* NOTDEF */

    /* process each node in the next event */
    while (list_val(id, (void **)&np) > 0) {
	if (np->type != LDDL_GROUP) {
	    if (do_normal_node(np)) {
		/* EOF */
		return 1;
	    }
	} else {
	    /* group - determine count */
	    if (np->cnt) {
		count = np->cnt;
	    } else if (np->nbytes_var != NULL) {
		var_getval(np->nbytes_var, &count);
	    } else {
		fatal("do_node_list: no count or nbytes_var");
	    }
	    if (_groupf != NULL) {
		(void)_groupf(np->argc, np->argv, np->group_name, Cur_evt_name);
	    }
	    if (np->argc && !Force) {
		Force++;
		Force_argc = np->argc;
		Force_argv = np->argv;
	    }
	    for (i=0; i<count; i++) {
		if (do_node_list(np->group_id, np->group_name)) {
		    /* EOF */
		    return 1;
		}
	    }
	    if (np->argc && !Force) {
		Force--;
	    }
	    if (_post_groupf != NULL) {
		(void)_post_groupf(np->argc, np->argv,
		    np->group_name, Cur_evt_name);
	    }
	}

    }
    /* end of event */

    return 0;
}

static int
do_normal_node(node *np)
{
    int argc = 0;
    char **argv = NULL;
    u_char bval;
    int count;
    double dval;
    float fval;
    int i;
    u_long lval;
    int nbytes;
    int ret;
    int stop;		/* used to back out of 'optional' items not present */
    u_short sval;

    /* sanity check */
    if (np->type == LDDL_GROUP) {
	fatal("do_normal_node: got a non-normal node!");
    }

    /* determine the size in bytes of this node */
    if (np->nbytes) {
	/* size from nbytes element */
	nbytes = np->nbytes;
    } else if (np->nbytes_var != NULL) {
	/* size from internal variable */
	var_getval(np->nbytes_var, &nbytes);
    } else {
	/* should not happen */
	fatal("%s: neither nbytes nor nbytes_var\n. Exiting.", Progname);
    }

    if (check_enough_data(nbytes)) {
	/* EOF */
	return 1;
    }

    switch (np->type) {
	case LDDL_INT16:
	    if (np->cnt) {
		count = np->cnt;
	    } else {
		count = nbytes / sizeof(short);
	    }

	    swab(Inp, Swapp, nbytes);

	    if (np->value_to_var) {
		/* value must be assigned to a variable */
		if (count != 1) {
		    warning("%s: warning: item 1 to variable (%ld)",
			Progname, Charcnt);
		}
		if (Native) {
		    memcpy(&sval, Inp, sizeof(short));
		} else {
		    memcpy(&sval, Swapp, sizeof(short));
		}
		var_setval(np->value_to_var, (int)sval);
	    }

	    stop = count;
	    if (np->value.is_set) {
		/* must be equal to a certain value */
		for (i=0; i < count; i++) {
		    if (Native) {
			memcpy(&sval, (u_short *)Inp + i, sizeof(short));
		    } else {
			memcpy(&sval, (u_short *)Swapp + i, sizeof(short));
		    }
		    if (np->mask.is_set) {
			sval &= np->mask.u.s;
		    }
		    if (sval != np->value.u.s) {
			if (np->optional) {
			    stop = i;
			    break;
			} else {
			    warning("%s: warning: (%04x) != %04x (%ld)",
				Progname, sval, np->value.u.s, Charcnt);
			}
		    }
		}
	    }

	    nbytes = stop * sizeof(short);

	    if (nbytes) {
		if (_nodef != NULL) {
		    if (np->argc) {
			argc = np->argc;
			argv = np->argv;
		    } else if (Force) {
			argc = Force_argc;
			argv = Force_argv;
		    }
		    /* node function */
		    (void)_nodef(Inp, Swapp, np->type, stop, nbytes, 
			argc, argv, Native, np->group_name, Cur_evt_name);
		}
	    }

	    break;

	case LDDL_INT8:
	    if (np->cnt) {
		count = np->cnt;
	    } else {
		count = nbytes;
	    }

	    memcpy(Swapp, Inp, nbytes);

	    if (np->value_to_var) {
		/* value must be assigned to a variable */
		if (count != 1) {
		    warning("%s: warning: item 1 to variable (%ld)",
			Progname, Charcnt);
		}
		var_setval(np->value_to_var, (int)(*Inp));
	    }

	    stop = count;
	    if (np->value.is_set) {
		/* must be equal to a certain value */
		for (i=0; i < count; i++) {
		    bval = *(Inp + i);
		    if (np->mask.is_set) {
			bval &= np->mask.u.b;
		    }
		    if (bval != np->value.u.b) {
			if (np->optional) {
			    stop = i;
			    break;
			} else {
			    warning("%s: warning: (%02x) != %02x (%ld)",
				Progname, bval, np->value.u.b, Charcnt);
			}
		    }
		}
	    }

	    nbytes = stop;

	    if (nbytes) {
		if (_nodef != NULL) {
		    if (np->argc) {
			argc = np->argc;
			argv = np->argv;
		    } else if (Force) {
			argc = Force_argc;
			argv = Force_argv;
		    }
		    /* node function */
		    (void)_nodef(Inp, Swapp, np->type, stop, nbytes, 
			argc, argv, Native, np->group_name, Cur_evt_name);
		}
	    }

	    break;
	case LDDL_INT32:
	    if (np->cnt) {
		count = np->cnt;
	    } else {
		count = nbytes / sizeof(long);
	    }

	    revbytes(Inp, Swapp, 4, count);

	    if (np->value_to_var) {
		/* value must be assigned to a variable */
		if (count != 1) {
		    warning("%s: warning: item 1 to variable (%ld)",
			Progname, Charcnt);
		}
		if (Native) {
		    memcpy(&lval, (u_long *)Inp, sizeof(long));
		} else {
		    memcpy(&lval, (u_long *)Swapp, sizeof(long));
		}
		var_setval(np->value_to_var, (int)lval);
	    }

	    stop = count;
	    if (np->value.is_set) {
		/* must be equal to a certain value */
		for (i=0; i < count; i++) {
		    if (Native) {
			memcpy(&lval, (u_long *)Inp + i, sizeof(long));
		    } else {
			memcpy(&lval, (u_long *)Swapp + i, sizeof(long));
		    }
		    if (np->mask.is_set) {
			lval &= np->mask.u.l;
		    }
		    if (lval != np->value.u.l) {
			if (np->optional) {
			    stop = i;
			    break;
			} else {
			    warning("%s: warning:node (%08x) != %08x (%ld)",
				Progname, lval, np->value.u.l, Charcnt);
			}
		    }
		}
	    }

	    nbytes = stop * sizeof(u_long);

	    if (nbytes) {
		if (_nodef != NULL) {
		    if (np->argc) {
			argc = np->argc;
			argv = np->argv;
		    } else if (Force) {
			argc = Force_argc;
			argv = Force_argv;
		    }
		    /* node function */
		    (void)_nodef(Inp, Swapp, np->type, stop, nbytes, 
			argc, argv, Native, np->group_name, Cur_evt_name);
		}
	    }

	    break;
	case LDDL_FLOAT32:
	    if (np->cnt) {
		count = np->cnt;
	    } else {
		count = nbytes / sizeof(float);
	    }

	    revbytes(Inp, Swapp, 4, count);

	    if (np->value_to_var) {
		/* value must be assigned to a variable */
		/* We don't do this for floats. */
		warning("%s: warning: can't assign float to var (%ld)",
		    Progname, Charcnt);
	    }

	    stop = count;
	    if (np->value.is_set) {
		/* must be equal to a certain value */
		for (i=0; i < count; i++) {
		    if (Native) {
			fval = *((float *)Inp + i);
		    } else {
			fval = *((float *)Swapp + i);
		    }
		    if (fval != np->value.u.f) {
			if (np->optional) {
			    stop = i;
			    break;
			} else {
			    warning("%s: warning: (%f) != %f (%ld)",
				Progname, fval, np->value.u.f, Charcnt);
			}
		    }
		}
	    }

	    nbytes = stop * sizeof(float);

	    if (nbytes) {
		if (_nodef != NULL) {
		    if (np->argc) {
			argc = np->argc;
			argv = np->argv;
		    } else if (Force) {
			argc = Force_argc;
			argv = Force_argv;
		    }
		    /* node function */
		    (void)_nodef(Inp, Swapp, np->type, stop, nbytes, 
			argc, argv, Native, np->group_name, Cur_evt_name);
		}
	    }

	    break;
	case LDDL_FLOAT64:
	    if (np->cnt) {
		count = np->cnt;
	    } else {
		count = nbytes / sizeof(double);
	    }

	    revbytes(Inp, Swapp, 8, count);

	    if (np->value_to_var) {
		/* value must be assigned to a variable */
		/* We don't do this for doubles. */
		warning("%s: warning: can't assign double to var (%ld)",
		    Progname, Charcnt);
	    }

	    stop = count;
	    if (np->value.is_set) {
		/* must be equal to a certain value */
		for (i=0; i < count; i++) {
		    if (Native) {
			dval = *((double *)Inp + i);
		    } else {
			dval = *((double *)Swapp + i);
		    }
		    if (dval != np->value.u.d) {
			if (np->optional) {
			    stop = i;
			    break;
			} else {
			    warning("%s: warning: (%f) != %f (%ld)",
				Progname, dval, np->value.u.d, Charcnt);
			}
		    }
		}
	    }

	    nbytes = stop * sizeof(double);

	    if (nbytes) {
		if (_nodef != NULL) {
		    if (np->argc) {
			argc = np->argc;
			argv = np->argv;
		    } else if (Force) {
			argc = Force_argc;
			argv = Force_argv;
		    }
		    /* node function */
		    (void)_nodef(Inp, Swapp, np->type, stop, nbytes, 
			argc, argv, Native, np->group_name, Cur_evt_name);
		}
	    }

	    break;
	default:
	    fatal("do_normal_node: unknown node type (%d)", np->type);
	    break;
    }
    /* end of a node */
    Inp += nbytes;
    Swapp += nbytes;
    Charcnt += nbytes;
    return 0;
}

int
yyerror(char *s)
{
    warning("%s near line %d\n", s, lineno_get());
}

void
newspec(int type, int cnt, char *nbytes_var, char *value_to_var, 
	int optional, valstruct *v, int mask, int group_id, char *group_name)
{
    node *np;
    
    np = fill_node(type, cnt, nbytes_var, value_to_var, optional,
    	v, mask, group_id, group_name);

    if (list_add(Cur_node_id, np)) {
	fatal("newspec: bad list_add");
    }
}

/* revbytes - reverse the bytes. For example, for width = 4 and cnt = 3:
 * abcdabcdabcd -> dcbadcbadcba. For width = 8 and cnt = 2:
 * abdcefghabcdefgh -> hgfedcbahgfedcba. This seems to be the effect we
 * want when converting longs, floats, and doubles. This may be too
 * simplistic to be correct. */
static void
revbytes(u_char *srcp, u_char *destp, int width, int cnt)
{
    u_char *dp;
    int n;
    u_char *sp;
    int w;
    int W;

    W = width - 1;
    sp = srcp;
    dp = destp;

    for (n=0; n < cnt; n++) {
	for (w=0; w < width; w++) {
	    dp[w] = sp[W-w];
	}
	sp += width;
	dp += width;
    }
}

/* group_init - start accumulating items in a group. There can only be
 * one group accumulating at a time. */
void
group_init(char *name)
{
    int ret;
    struct groupnode *sp;

    if (group_defined(name) != -1) {
	fatal("group_init: group '%s' already in use", name);
    }

    sp = emalloc(sizeof(groupnode));
    sp->name = str_alloc(NULL, name);

    sp->id = list_new();
    if (sp->id < 0) {
	fatal("group_init: can't make new list");
    }
    strcpy(sp->name, name);
    Cur_group_id = sp->id;
    Cur_group_name = sp->name;
    ret = list_add(Group_list_id, sp);
    if (ret) {
	fatal("group_init: bad list_add");
    }
}

/* group_add - add an item to the open group */
void
group_add(int type, int cnt, char *nbytes_var, char *value_to_var, 
	int optional, valstruct *v, int mask)
{
    node *np;
    int ret;

    if (Cur_group_id == -1) {
	fatal("group_add: not defining group");
    }

    np = fill_node(type, cnt, nbytes_var, value_to_var,
	optional, v, mask, Cur_group_id, Cur_group_name);

    /* Put it on the linked list. */
    ret = list_add(Cur_group_id, np);
    if (ret) {
	fatal("group_add: bad list_add");
    }
}

/* group_end - stop accumulating items in the open group */
void
group_end(void)
{
    Cur_group_id = -1;
}

/* group_defined - see if 'name' corresponds to a defined group. If so,
 * return its "id", else return -1. */
int
group_defined(char *name)
{
    struct groupnode *sp;

    if (list_first(Group_list_id)) {
	return -1;
    }

    while (list_val(Group_list_id, (void **)&sp) > 0) {
	if (STREQ(sp->name, name)) {
	    return sp->id;
	}
    }

    return -1;
}

/* evt_def - get ready to start adding the nodes for a particular evt kind */
void 
evt_def(char *name)
{
    evt *ep;
    int got = 0;

    /* Find the id of the node list for this kind of event */
    if (list_first(Evt_kind_id)) {
	fatal("evt_def: bad list_first");
    }

    while (list_val(Evt_kind_id, (void **)&ep) > 0) {
	if (STREQ(name, ep->name)) {
	    got = 1;
	    break;
	}
    }

    if (!got) {
	fatal("evt_def: no such event kind '%s'", name);
    }

    ep->argc = args_cur_argc();
    ep->argv = args_cur_argv();
    args_reset_cur_argc();

    Cur_node_id = ep->node_id;
    Cur_evt_name = str_alloc(Cur_evt_name, name);
    strcpy(Cur_evt_name, name);
    Check_evt_kind = 1;
}

/* evt_kind_init - start defining a new event kind */
void 
evt_kind_init(char *name)
{
    evt *ep;
    
    ep = emalloc(sizeof(evt));
    ep->name = str_alloc(NULL, name);
    strcpy(ep->name, name);

    ep->cond_id = list_new();
    if (ep->cond_id < 0) {
	fatal("evt_kind_init: bad cond list_new");
    }
    ep->node_id = list_new();
    if (ep->node_id < 0) {
	fatal("evt_kind_init: bad node list_new");
    }
    ep->argc = 0;
    ep->argv = NULL;
    if (list_add(Evt_kind_id, ep)) {
	fatal("evt_kind_init: bad list_add");
    }
    Cur_cond_id = ep->cond_id;
}

/* evt_kind_value - record the value and name of a particular evt kind */
void 
evt_kind_value(int offset, int type, int value, int mask)
{
    evt_kind_cond *cp;
    int r;
    
    cp = emalloc(sizeof(evt_kind_cond));
    cp->offset = offset;
    cp->type = type;

    if (type == LDDL_INT8) {
	cp->val.u.b = (unsigned char)value;
	r = offset + sizeof(char);
	if (mask) {
	    cp->mask.u.b = (unsigned char)mask;
	}
    } else if (type == LDDL_INT16) {
	cp->val.u.s = (unsigned short)value;
	r = offset + sizeof(short);
	if (mask) {
	    cp->mask.u.s = (unsigned short)mask;
	}
    } else if (type == LDDL_INT32) {
	cp->val.u.l = (unsigned long)value;
	r = offset + sizeof(long);
	if (mask) {
	    cp->mask.u.l = (unsigned long)mask;
	}
    }

    if (mask) {
	cp->mask.is_set = 1;
    } else {
	cp->mask.is_set = 0;
    }

    if (r > Evt_kind_rdamt) {
	Evt_kind_rdamt = r;
    }

    if (list_add(Cur_cond_id, cp)) {
	fatal("evt_kind_value: bad list_add");
    }
}

/* check_enough_data - returns 1 on EOF, else 0 */
static int
check_enough_data(int atleast)
{
    static int Cur_buf_size;	/* size in bytes of data buffers */
    static u_char *D;		/* input data buffer */
    static u_char *Dend;	/* end of input data buffer */
    static u_char *Dswap;	/* swapped data buffer */
    int diff;
    static int First_time = 1;
    static int More_data = 1;
    size_t nwrit;
    int ret;

    if (First_time) {
	/* initialize buffers and do first read */
	First_time = 0;
	D = emalloc(DEF_BUF_SIZE);
	Cur_buf_size = DEF_BUF_SIZE;
	Dend = D + Cur_buf_size;
	Inp = D;
	
	Dswap = emalloc(DEF_BUF_SIZE);
	Swapp = Dswap;

	ret = getbuf(D, sizeof(char), Cur_buf_size);
	if (ret != Cur_buf_size) {
	    /* EOF */
	    Cur_buf_size = ret;
	    Dend = D + Cur_buf_size;
	    More_data = 0;
	} else {
	    More_data = 1;
	}
    }

    if (Inp + atleast <= Dend) {
	/* enough data already in buf */
	return 0;
    }

    if (!More_data) {
	/* already at EOF */
	return 1;
    }

    diff = Dend - Inp;		/* amt of unprocessed data */
    memmove(D, Inp, diff);	/* put unprocessed data at beginning of buf */

    if (atleast > Cur_buf_size - diff) {
	/* requesting more data than we have room for */
	Cur_buf_size = atleast + diff;
	D = erealloc(D, Cur_buf_size);
	Dend = D + Cur_buf_size;
	Dswap = erealloc(Dswap, Cur_buf_size);
    }

    ret = getbuf(D+diff, sizeof(char), Cur_buf_size - diff);
    if (ret != Cur_buf_size - diff) {
	More_data = 0;
	Cur_buf_size = ret + diff;
	Dend = D + Cur_buf_size;
    }

    Inp = D;
    Swapp = Dswap;
    return 0;
}

static int
match_evt_kind(evt_kind_cond *cp, u_char *p)
{
    u_char bval;
    u_long lval;
    int ret;
    u_short sval;

    ret = 0;

    if (cp->type == LDDL_INT8) {
	bval = *p;
	if (cp->mask.is_set) {
	    bval &= cp->mask.u.b;
	}
	if (bval == cp->val.u.b) {
	    ret = 1;
	}
    } else if (cp->type == LDDL_INT16) {
	if (Native) {
	    memcpy(&sval, p, sizeof(short));
	} else {
	    swab(p, (char *)&sval, sizeof(short));
	}
	if (cp->mask.is_set) {
	    sval &= cp->mask.u.s;
	}
	if (sval == cp->val.u.s) {
	    ret = 1;
	}
    } else if (cp->type == LDDL_INT32) {
	if (Native) {
	    memcpy(&lval, p, sizeof(long));
	} else {
	    revbytes(p, (u_char *)&lval, sizeof(long), 1);
	}
	if (cp->mask.is_set) {
	    lval &= cp->mask.u.l;
	}
	if (lval == cp->val.u.l) {
	    ret = 1;
	}
    } else {
	fatal("match_evt_kind: bad type");
    }

    return ret;
}

static int Depth = 0;

static void
print_node(node *np)
{
    char prefix[1024];
    char valstr[32];
    char maskstr[32];

    memset(prefix, '\t', Depth);
    prefix[Depth+1] = '\0';

    if (np == NULL) {
	fprintf(stderr, "%s NULL\n", prefix);
	return;
    }

    sprintf(valstr, "----");

    fprintf(stderr, "%s type ", prefix);
    if (np->type == LDDL_INT8) {
	fprintf(stderr, "byte\n");
	if (np->value.is_set) {
	    sprintf(valstr, "%u (%02x)", np->value.u.b, np->value.u.b);
	}
	if (np->mask.is_set) {
	    sprintf(maskstr, "%02x", np->value.u.b);
	}
    } else if (np->type == LDDL_INT16) {
	fprintf(stderr, "int\n");
	if (np->value.is_set) {
	    sprintf(valstr, "%u (%04x)", np->value.u.s, np->value.u.s);
	}
	if (np->mask.is_set) {
	    sprintf(maskstr, "%04x", np->value.u.s);
	}
    } else if (np->type == LDDL_INT32) {
	fprintf(stderr, "long\n");
	if (np->value.is_set) {
	    sprintf(valstr, "%ld (%08x)", np->value.u.l, np->value.u.l);
	}
	if (np->mask.is_set) {
	    sprintf(maskstr, "%08x", np->value.u.l);
	}
    } else if (np->type == LDDL_FLOAT32) {
	fprintf(stderr, "float\n");
	if (np->value.is_set) {
	    sprintf(valstr, "%f (%08x)", np->value.u.f, np->value.u.f);
	}
	if (np->mask.is_set) {
	    sprintf(maskstr, "%08x", np->value.u.f);
	}
    } else if (np->type == LDDL_FLOAT64) {
	fprintf(stderr, "double\n");
	if (np->value.is_set) {
	    sprintf(valstr, "%f (%016x)", np->value.u.d, np->value.u.d);
	}
	if (np->mask.is_set) {
	    sprintf(maskstr, "%016x", np->value.u.d);
	}
    } else if (np->type == LDDL_GROUP) {
	if (np->group_name == NULL) {
	    fprintf(stderr, "group ()\n");
	} else {
	    fprintf(stderr, "group (%s)\n",np->group_name);
	}
	if (np->value.is_set) {
	    sprintf(valstr, "%u (%04x)", np->value.u.s, np->value.u.s);
	}
	if (np->mask.is_set) {
	    sprintf(maskstr, "%04x", np->value.u.s);
	}
    } else {
	fprintf(stderr, "unknown\n");
    }

    fprintf(stderr, "%s count %d\n", prefix, np->cnt);
    fprintf(stderr, "%s nbytes %d\n", prefix, np->nbytes);
    if (np->nbytes_var) {
	fprintf(stderr, "%s nbytes_var '%s'\n", prefix, np->nbytes_var);
    }
    if (np->value_to_var) {
	fprintf(stderr, "%s value_to_var '%s'\n", prefix, np->value_to_var);
    }
    fprintf(stderr, "%s optional %d\n", prefix, np->optional);
    fprintf(stderr, "%s group_id %d\n", prefix, np->group_id);
    fprintf(stderr, "%s value %s\n", prefix, valstr);
    fprintf(stderr, "%s mask %s\n", prefix, maskstr);
    fprintf(stderr, "%s \n", prefix);
}

void
print_evts (void)
{
    evt *ep;
    int ret;

    if (list_first(Evt_kind_id)) {
	fatal("print_evts: bad list_first(Evt_kind_id)");
    }

    while (list_val(Evt_kind_id, (void **)&ep) > 0) {
	fprintf(stderr, "%s\n", ep->name);
	fprintf(stderr, "cond_id %d\n", ep->cond_id);
	fprintf(stderr, "node_id %d\n", ep->node_id);
	print_node_list(ep->node_id);
    }
}

static void
print_node_list(int id)
{
    int nodecnt;
    node *np;
    char prefix[1024];

    fprintf(stderr, "%s list id %d\n", prefix, id);

    Depth++;
    memset(prefix, '\t', Depth);
    prefix[Depth+1] = '\0';

    if (list_first(id)) {
	fprintf(stderr, "%s empty\n", prefix);
	Depth--;
	return;
    }

    while (list_val(id, (void **)&np) > 0) {
	print_node(np);
	if (np->type == LDDL_GROUP) {
	    print_node_list(np->group_id);
	}
    }

    Depth--;
}

static node *
fill_node(int type, int cnt, char *nbytes_var, char *value_to_var, 
	int optional, valstruct *v, int mask, int group_id,
	char *group_name)
{
    node *np;
    char *p;

    np = emalloc(sizeof(node));
    np->type = type;
    np->cnt = cnt;
    np->group_id = group_id;
    if (group_name == NULL) {
	np->group_name = NULL;
    } else {
	np->group_name = str_alloc(NULL, group_name);
	strcpy(np->group_name, group_name);
    }

    if (mask) {
	np->mask.is_set = 1;
    } else {
	np->mask.is_set = 0;
    }

    if (np->type == LDDL_INT8) {
	np->nbytes = np->cnt * sizeof(char);
	if (mask) {
	    np->mask.u.b = (u_char)mask;
	}
    } else if (np->type == LDDL_INT16) {
	np->nbytes = np->cnt * sizeof(short);
	if (mask) {
	    np->mask.u.s = (u_short)mask;
	}
    } else if (np->type == LDDL_INT32) {
	np->nbytes = np->cnt * sizeof(long);
	if (mask) {
	    np->mask.u.l = (u_long)mask;
	}
    } else if (np->type == LDDL_FLOAT32) {
	np->nbytes = np->cnt * sizeof(float);
    } else if (np->type == LDDL_FLOAT64) {
	np->nbytes = np->cnt * sizeof(double);
    } else if (np->type == LDDL_GROUP) {
	/* group */
	np->nbytes = -1;
    }

    if (nbytes_var == NULL) {
	np->nbytes_var = NULL;
    } else {
	np->nbytes = 0;
	np->nbytes_var = str_alloc(NULL, nbytes_var);
	strcpy(np->nbytes_var, nbytes_var);
	p = strchr(np->nbytes_var, '\n');
	if (p != NULL) {
	    *p = '\0';
	}
    }

    if (value_to_var == NULL) {
	np->value_to_var = NULL;
    } else {
	np->value_to_var = str_alloc(NULL, value_to_var);
	strcpy(np->value_to_var, value_to_var);
	p = strchr(np->value_to_var, '\n');
	if (p != NULL) {
	    *p = '\0';
	}
    }
    
    if (v == NULL) {
	np->value.is_set = 0;
	if (optional) {
	    warning("fill_node: optional flag ignored - no value specified");
	    optional = 0;
	}
    } else {
	memcpy(&(np->value), v, sizeof(valstruct));
	np->value.is_set = 1;
    }
    
    np->optional = optional;

    np->argc = args_cur_argc();
    np->argv = args_cur_argv();
    args_reset_cur_argc();

    return np;
}

void
lddl_set_pre_evt_func(evt_cb f)
{
    _pre_evtf = f;
}

void 
lddl_set_group_func(group_cb f)
{
    _groupf = f;
}

void 
lddl_set_post_group_func(group_cb f)
{
    _post_groupf = f;
}

void
lddl_set_evt_func(evt_cb f)
{
    _evtf = f;
}

void
lddl_set_post_evt_func(evt_cb f)
{
    _post_evtf = f;
}

void
lddl_set_node_func(node_cb f)
{
    _nodef = f;
}

void
lddl_set_err_func(void (*f)(char *str))
{
    _app_errf = f;
}

void
lddl_set_native(int native)
{
    Native = native;
    Native_set = 1;
}

char *
lddl_version(void)
{
    return LDDL_VERSION;
}

void 
fatal(char *fmt, ...)
{
    va_list ap;
    char str[1024];

    va_start(ap, fmt);
    vsnprintf(str, 1024, fmt, ap);
    if (_app_errf != NULL) {
	_app_errf(str);
    } else {
	fprintf(stderr, "%s\n", str);
	exit(1);
    }
}

void 
warning(char *fmt, ...)
{
    va_list ap;
    char str[1024];

    va_start(ap, fmt);
    vsnprintf(str, 1024, fmt, ap);
    fprintf(stderr, "%s\n", str);
}

/* endian_determine - figure out from the datastream if the data is in our
 * system's native endianness. Returns 1 on EOF, else 0. */
static int
endian_determine(void)
{
    u_char *cp;
    u_long lval;
    u_short sval;

    if (check_enough_data(Endian_rdamt)) {
	/* EOF */
	return 1;
    }

    cp = Inp + Endian_offset;

    if (Endian_type == LDDL_INT16) {
	memcpy(&sval, cp, sizeof(short));
	if (Endian_mask.is_set) {
	    Native = (Endian_val.u.s == (sval & Endian_mask.u.s));
	} else {
	    Native = (Endian_val.u.s == sval);
	}
    } else if (Endian_type == LDDL_INT32) {
	memcpy(&lval, cp, sizeof(long));
	if (Endian_mask.is_set) {
	    Native = (Endian_val.u.l == (lval & Endian_mask.u.l));
	} else {
	    Native = (Endian_val.u.l == lval);
	}
    } else {
	/* incorrect type - shouldn't happen */
	warning( "endian_determine: incorrect Endian_type = %d",
	    Endian_type);
    }

    return 0;
}

void
endian_set(int offset, int type, int value, int mask)
{
    if (Native_set) {
	warning(
	    "endian_set: Native set (%d) on cmd line overrides format file",
	    Native);
	return;
    }

    Check_endianness = 0;

    if (offset < 0) {
	warning("endian_set: offset must be > 0 (not %d)", offset);
	return;
    }

    if (type == LDDL_INT16) {
	Endian_val.is_set = 1;
	Endian_val.u.s = (u_short)value;
	if (mask) {
	    Endian_mask.is_set = 1;
	    Endian_mask.u.s = (u_short)mask;
	}
	Endian_rdamt = offset + sizeof(short);
    } else if (type == LDDL_INT32) {
	Endian_val.is_set = 1;
	Endian_val.u.l = (u_long)value;
	if (mask) {
	    Endian_mask.is_set = 1;
	    Endian_mask.u.l = (u_long)mask;
	}
	Endian_rdamt = offset + sizeof(long);
    } else {
	warning("endian_set: bad endian type. Only use I or L.");
	return;
    }

    Endian_type = type;
    Endian_offset = offset;
    Check_endianness = 1;
}
