/* scan.l - flex lexical scanner */

%{
#include <stdlib.h>	/* atof */
#include <string.h>	/* strchr, strdup */
#include "y.tab.h"
#include "misc.h"	/* lineno_incr, lineno_get */

static char *cvsid = "$Id: scan.l,v 1.1.1.1 2000/03/24 20:10:24 marty Exp $";

static void do_escape_chars(void);

/* For debugging output, set Scan_debug to a nonzero value. */
static int Scan_debug = 0;
#define F(x)	((Scan_debug) ? fprintf(stderr,"\t\t%s (%s)\n", (x), yytext) : 1)
%}

NUMBER		[0-9]+
FLOATVAL	[0-9]+"."[0-9]*
HEXVAL		"0x"[a-fA-F0-9]+
ID		[a-zA-Z_][a-zA-Z_0-9]*

%%
[ \t]+	/* eat white space */ F("space");

^#.*\n	{
	    /* Comment at the beginning of a line: return nothing */
	    F("comment begin-line");
	    lineno_incr();
	    /* no return */
}

#.*\n+	{
	    /* Comments in middle of a line: return a '\n' after slurping
	     * up any adjacent newlines. */
	    int i;
	    char *p;
	    F("comment mid-line");
	    p = strchr(yytext, '\n');
	    for (i=0; i<strlen(p); i++) {
		lineno_incr();
	    }
	    return '\n';
}


"sig"		{ F("sig");		return SIG; }
"evt"		{ F("evt");		return EVT; }
"endian"	{ F("endian");		return ENDIAN; }
"mask"		{ F("mask");		return MASK; }
"offset"	{ F("offset");		return OFFSET; }
"optional"	{ F("optional");	return OPTIONAL; }
"<print_evts>"	{ F("<print_evts>");	return PRINT_EVTS; }
"group_end"	{ F("group_end");	return GROUPEND; }
"group"		{ F("group");		return GROUP; }
"type"		{ F("type");		return TYPE; }

"&&"		{ F("&&");		return AND; }
"->"		{ F("->");		return ARROW; }
=		{ F("=");		return '='; }

I		{ F("I");		return T_SHORT; }
S		{ F("S");		return T_SHORT; }
L		{ F("L");		return T_LONG; }
B		{ F("B");		return T_BYTE; }
F		{ F("F");		return T_FLOAT; }
D		{ F("D");		return T_DOUBLE; }
:		{ F("SEP");		return SEPARATOR; }

{NUMBER}	{
	    /* number */
	    F("NUMBER");
	    yylval.ival = atoi(yytext);
	    return NUM;
}

{HEXVAL}	{
	    /* hex number */
	    F("HEXVAL");
	    yylval.ival = strtol(yytext, (char **)NULL, 16);
	    return NUM;
}

{FLOATVAL}	{
	    /* floating pt value */
	    F("FLOATVAL");
	    yylval.fval = atof(yytext);
	    return FLOATNUM;
}

{ID}	{
	    /* string */
	    F("ID");
	    yylval.str = strdup(yytext);
	    return STRING;
}

\n+	{
	    int i;
	    F("newline");
	    for (i=0; i<strlen(yytext); i++) {
		lineno_incr();
	    }
	    return '\n';
}

\"[^\"]*\"	{
	    /* anything in quotes */
	    F("FMT");
	    do_escape_chars();
	    yytext[yyleng-1] = '\0';	/* get rid of end quotes */
	    yylval.str = strdup(yytext+1);	/* get rid of start quotes */
	    return STRING;
}

[^ \t\n]*	{
	    /* unknown string - may be a user-defined keywd or arg */
	    F("UNK");
	    do_escape_chars();
	    yylval.str = strdup(yytext);
	    return STRING;

}
%%
static void
do_escape_chars(void)
{
    char c;
    int i;

    for (i=0; i<yyleng; i++) {
	if (yytext[i] == '\\') {
	    switch(yytext[i+1]) {
		case 'a':
		    c = '\a';
		    break;
		case 'b':
		    c = '\b';
		    break;
		case 'f':
		    c = '\f';
		    break;
		case 'n':
		    c = '\n';
		    break;
		case 't':
		    c = '\t';
		    break;
		case 'v':
		    c = '\v';
		    break;
		default:
		    c = '\\';	/* ??? */
		    break;
	    }
	    yytext[i] = c;
	    memmove(yytext+i+1, yytext+i+2, yyleng-(i+2));
	    yyleng--;
	}
    }
}
