/* proc.c - generic data analysis framework using lddl library
 *
 * NOTE: put your data analysis code in the function 'process_evt' below.
 *
 * Marty Olevitch, Jan 01
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <lddl.h>
#include "data.h"

static char *cvsid = "$Id: proc.c,v 1.2 2001/03/07 19:57:50 marty Exp $";

#ifndef STREQ
#define STREQ(a, b)	(strcmp((a), (b)) == 0)
#endif

static int evt(char *name, int argc, char **argv);
static void fill_evt_struct(struct realevt *r);
static int group(int argc, char **argv, char *grpname, char *evtname);
static int node(char *inp, char *swapp, int type, int count, int nbytes, int
    argc, char **argv, int native, char *group_name, char *evt_name);
static int post_evt(char *name, int argc, char **argv);
static void process_evt(void);

static void usage(void);

static char *Curp;	/* ptr to current location in Data buffer */
static char *Data;	/* ptr to start of event's accumulated data */
static unsigned long Evtbytes = 0L;
static unsigned long Evtcnt = 0L;
static unsigned long Evtno = 0L;
static int Evt_overflow;	/* nonzero if too much data in evt */
static const long MAXDATA = 260000L;	/* max number of bytes in any event */
static char *Progname;

int
main(int argc, char *argv[])
{
    int c;
    int errflg = 0;
    char *fmtfile = NULL;
    extern int optind;
    extern char *optarg;

    Progname = argv[0];

    while ((c = getopt(argc, argv, "f:Nn")) != EOF) {
	switch (c) {
	    case 'f':
	    	fmtfile = optarg;
		break;
	    case 'N':
	    	/* assume native endianness */
		lddl_set_native(1);
		break;
	    case 'n':
	    	/* assume non-native endianness */
		lddl_set_native(0);
		break;
	    case '?':
	    default:
	    	errflg++;
		break;
	}
    }

    if (errflg) {
	usage();
	exit(1);
    }

    Data = (char *)malloc(MAXDATA);
    if (Data == NULL) {
	fprintf(stderr, "%s: can't malloc (%s). Exiting.\n",
	    Progname, strerror(errno));
	exit(1);
    }

    lddl_set_group_func(group);
    lddl_set_evt_func(evt);
    lddl_set_post_evt_func(post_evt);
    lddl_set_node_func(node);
    (void)lddl_search_for_evt_start(1);
    lddl_start(fmtfile, argc-optind, argv+optind, Progname);
    /* printf("%d\n", Evtcnt); */
    return 0;
}

/* evt - start accumulating the data of a new event */
static int 
evt(char *name, int argc, char **argv)
{
    Curp = Data;
    Evtbytes = 0L;
    Evt_overflow = 0;
    return 0;
}

/* post_evt - process the accumulated data of the event */
static int 
post_evt(char *name, int argc, char **argv)
{
    if (Evt_overflow) {
	fprintf(stderr, "%s: event overflow\n", Progname);
	return 0;
    }
    process_evt();
    return 0;
}

/* group - no-op for now */
static int
group(int argc, char **argv, char *group_name, char *evtname)
{
    return 0;
}

/* node - just copy the data */
static int
node(char *inp, char *swapp, int type, int count, int nbytes, int
    argc, char **argv, int native, char *group_name, char *evt_name)
{
    char *p;

    if (Evtbytes + nbytes > MAXDATA) {
	Evt_overflow = 1;
    }
    if (Evt_overflow) {
	return 0;
    }
    p = native ? inp : swapp;
    memcpy(Curp, p, nbytes);
    Curp += nbytes;
    Evtbytes += nbytes;
    return 0;
}

static void
usage(void)
{
    fprintf(stderr,
    	"\nusage: %s -f fmtfile [-N] [-n] [-o outfile] [file ...]\n",
    	Progname);
    fprintf(stderr, "     -f fmtfile    name of format file\n");
    fprintf(stderr, "     -N            data IS in native endianness\n");
    fprintf(stderr, "     -n            data IS NOT in native endianness\n");
    fprintf(stderr, "     file          input data file(s) (def. stdin)\n");
}

static void
process_evt(void)
{
    unsigned char *cp;
    unsigned long *lp;
    unsigned short *sp;
    struct realevt r;

    Evtno++;

    /* type */
    sp = (unsigned short *)Data;
    sp += 2;
    if (*sp != TYPE_REAL) {
	/* Just mess with real events. */
	return;
    }

    fill_evt_struct(&r);

    /* Process evt here -- your code goes here! */
    fprintf(stderr, "%lu\n", r.std.abs_evtno);
}

static void
fill_evt_struct(struct realevt *r)
{
    unsigned char *cp;
    int i;
    unsigned long *lp;
    unsigned short *sp;

    sp = (unsigned short *)Data;
    sp += OFF_TYPE;

    r->std.type = *sp++;
    r->std.exp_id = *sp++;

    lp = (unsigned long *)sp;
    r->std.size = *lp++;
    r->std.evtno = *lp++;
    r->std.abs_evtno = *lp++;
    r->std.clockticks = *lp++;
    r->std.vidsize2 = *lp++;
    lp++;
    r->std.vidsize1 = *lp;

    sp = (unsigned short *)Data;
    sp += OFF_PHA;	/* start of pha stuff */

    for (i=0; i < NWUPHA; i++) {
	r->pha.wupha[i] = *sp++;
    }
    for (i=0; i < NMARSHALLPHA; i++) {
	r->pha.marshallpha[i] = *sp++;
    }

    sp = (unsigned short *)Data;
    sp += OFF_DISC;

    cp = (unsigned char *)sp;
    r->disc.cam_discrim = *cp++;
    r->disc.iigain = *cp++;

    sp = (unsigned short *)cp;
    r->disc.status = *sp++;
    for (i=0; i < NSCAL; i++) {
	r->disc.selscalar[i] = *sp++;
    }
    for (i=0; i < NSCAL; i++) {
	r->disc.sumscalar[i] = *sp++;
    }

    sp = (unsigned short *)Data;
    sp += OFF_VID;
    memcpy(r->vid, sp, LEN_VID * sizeof(short));
}
