
/*
 Copyright (C) 2003 Martin A. Olevitch
 Written by Martin A. Olevitch (marty@cosray.wustl.edu)

 This file is part of LDDL.

 LDDL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 LDDL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with LDDL, see the file COPYING; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* grammar.y - dataswap bison grammar */

%{
#include "lddl.h"	/* LDDL_INT8, LDDL_INT16, etc. */
#include "ds.h"
#include "misc.h"
#include "args.h"

#include <string.h>

static char *cvsid = "$Id: grammar.y,v 1.1.1.1 2000/03/24 20:10:24 marty Exp $";

static valstruct V;
static valstruct *Vptr = NULL;
static int Curtype;
static char *Group_name = NULL;
static int In_sig = 0;		/* if nonzero, processing a "signature" */

#define YYDEBUG 1
int yydebug = 0;
%}

%union {
char * str;	/* for STRING */
double fval;	/* for floats/doubles */
int ival;	/* for everything else */
}

%token <ival>	AND
%token <ival>	ARROW
%token <ival>	BITAND
%token <ival>	BRACE_LEFT
%token <ival>	BRACE_RIGHT
%token <ival>	ENDIAN		/* endian definition */
%token <ival>	EVT		/* evt definition */
%token <fval>	FLOATNUM	/* floating pt value */
%token <ival>	GROUP
%token <ival>	GROUPEND
%token <ival>	MASK
%token <ival>	NUM		/* integer value */
%token <ival>	OFFSET
%token <ival>	OPTIONAL
%token <ival>	PRINT_EVTS
%token <ival>	SHIFTR
%token <ival>	SIG		/* evt signature */
%token <str>	STRING		/* character string */
%token <ival>	SEPARATOR
%token <ival>	T_BYTE
%token <ival>	T_DOUBLE
%token <ival>	T_FLOAT
%token <ival>	T_SHORT
%token <ival>	T_LONG
%token <ival>	TYPE

%type <ival>	arg
%type <ival>	arglist
%type <ival>	args
%type <str>	assign_from
%type <str>	assign_to
%type <ival>	bitand
%type <ival>	cond
%type <ival>	cond_args
%type <ival>	count
%type <ival>	endian
%type <ival>	evt
%type <ival>	keep_start
%type <ival>	keep_end
%type <ival>	line
%type <ival>	mask
%type <ival>	number
%type <ival>	offset
%type <ival>	optional
%type <ival>	signature
%type <ival>	signature_start
%type <ival>	group_def_end
%type <ival>	group_def
%type <ival>	shiftr
%type <ival>	typespec
%type <ival>	type
%type <ival>	value

%%
input :	  /* empty */
	| input line
	| input '\n'
;

line :	  typespec
	| group_def
	| group_def_end
	| evt
	| signature
	| endian
	| keep_start
	| keep_end
	| PRINT_EVTS	{ print_evts(); }
;

group_def :	  GROUP STRING	{ group_init($2); }
;

group_def_end :  GROUPEND { group_end(); }
;

typespec: type count bitand shiftr value mask optional args '\n'	{
	    newspec($1, $2, NULL, NULL, $7, Vptr, $6, Group_name,
		 (unsigned long)$3, $4);
	  }
	| type count bitand shiftr assign_to args '\n' {
	    newspec($1, $2, NULL, $5, 0, NULL, 0, Group_name,
		 (unsigned long)$3, $4);
	  }
	| type count bitand shiftr assign_from args '\n' {
	    newspec($1, 0, $5, NULL, 0, NULL, 0, Group_name,
		 (unsigned long)$3, $4);
	  }
;

type :	  T_SHORT	{ $$ = LDDL_INT16; Curtype = $$; Group_name = NULL; }
	| T_LONG	{ $$ = LDDL_INT32; Curtype = $$; Group_name = NULL; }
	| T_BYTE	{ $$ = LDDL_INT8; Curtype = $$; Group_name = NULL; }
	| T_FLOAT	{ $$ = LDDL_FLOAT32; Curtype = $$; Group_name = NULL; }
	| T_DOUBLE	{ $$ = LDDL_FLOAT64; Curtype = $$; Group_name = NULL; }
	| STRING	{
	    /* insert group definition */
	    Group_name = str_alloc(Group_name, $1);
	    strcpy(Group_name, $1);
	    $$ = LDDL_GROUP;
	}
;

optional : /* empty */	{ $$ = 0; }
	| OPTIONAL	{ $$ = 1; }
;

count :	  /* empty */		{ $$ = 1; }	/* default val = 1 */
	| NUM 		{
			    if ($1 == 0) {
				fprintf(stderr, "count of zero near line %d\n",
				    lineno_get());
				return 1;
			    }
	}
;

value:	  /* empty */	{ Vptr = NULL; }
	| number	{ Vptr = &V; }
;

number:	  '=' NUM	{
		if (Curtype == LDDL_INT16) {
		    V.u.s = (unsigned short)$2;
		} else if (Curtype == LDDL_INT32) {
		    V.u.l = (unsigned long)$2;
		} else if (Curtype == LDDL_INT8) {
		    V.u.b = (unsigned char)$2;
		} else {
		    fprintf(stderr,
			"set non-integer type to integer number near line %d\n",
			lineno_get());
		    return 1;
		}
		V.is_set = 1;
	    }

	| '=' FLOATNUM	{
		if (Curtype == LDDL_FLOAT32) {
		    V.u.f = (float)$2;
		} else if (Curtype == LDDL_FLOAT64) {
		    V.u.d = (double)$2;
		} else {
		    fprintf(stderr,
			"set non-float type to float number near line %d\n",
			lineno_get());
		    return 1;
		}
		V.is_set = 1;
	    }
;

assign_to:	ARROW STRING	{ $$ = $2; }
;

assign_from:	STRING	{ $$ = $1; }
;

endian:		ENDIAN cond
;

signature:		signature_start cond	{ In_sig = 0; }
;

signature_start:	SIG STRING	{
			    In_sig = 1;
			    evt_kind_init($2);
    			}
;

cond:	  cond_args
	| cond_args AND cond
;

cond_args:  offset TYPE type '=' NUM mask {
		if (In_sig) {
		    evt_kind_value($1, $3, $5, $6);
		} else {
		    endian_set($1, $3, $5, $6);
		}
	  }
;

offset:	  /* empty */	{ $$ = 0; }
	| OFFSET NUM	{ $$ = $2; }
;

keep_start:	BRACE_LEFT	{ keep_start(); $$ = 1; }
;

keep_end:	BRACE_RIGHT	{ keep_end(); $$ = 1; }
;

mask:	  /* empty */	{ $$ = 0; } /* mask of 0 means "don't mask" */
	| MASK NUM	{ $$ = $2; }
;

bitand:	  /* empty */	{ $$ = 0; }
	| BITAND NUM	{ $$ = $2; }
;

shiftr:	  /* empty */	{ $$ = 0; }
	| SHIFTR NUM	{ $$ = $2; }
;

evt:	EVT STRING args	{ evt_def($2); }
;

args:	  /* empty */	{ $$ = 0; }
	| SEPARATOR arglist
;

arglist:  arg
	| arglist arg
;

arg:	  STRING	{ args_add($1); }
	| NUM
		{
		    char str[128];
		    snprintf(str, 128, "%d", $1);
		    args_add(str);
		}
	| FLOATNUM
		{
		    char str[128];
		    snprintf(str, 128, "%f", $1);
		    args_add(str);
		}
;
%%

/* mitten_drin ("in the middle of") - this routine can be called after
 * yyparse() has returned to see if we were in the middle of something when
 * we ran out of input. The function prototype is found in ds.h. */
int
mitten_drin(void)
{
    // Add an error message [fprintf(stderr,...] here and return 1 if there
    // is a problem.

    // Next line should be last.
    return 0;
}
